# -*- coding: utf-8 -*-


import numpy as np
import scipy.constants as scpct
import datastock as ds


from . import _utils_surface3d
from . import _class4_check

from ..geom._comp_solidangles import _check_polygon_2d


_DMAT_KEYS = {
    'name': {
        'types': str,
        'can_be_None': True,
    },
    'symbol': {
        'types': str,
        'can_be_None': True,
    },
    'thickness': {
        'types': float,
        'sign': '> 0.',
        'can_be_None': True,
    },
    'qeff_E': {
        'dtype': float,
        'sign': '> 0.',
        'can_be_None': True,
    },
    'qeff': {
        'dtype': float,
        'sign': ['>= 0.', '<= 1.'],
        'can_be_None': True,
    },
    'mode': {
        'types': str,
        'default': 'current',
        'allowed': ['current', 'PHA'],
        'can_be_None': False,
    },
    'bins': {
        'dtype': float,
        'sign': '>0.',
        'unique': True,
        'can_be_None': True,
    },
}


# #####################################################################
# #####################################################################
#                   Camera 1d
# #####################################################################


def _camera_1d_check(
    coll=None,
    key=None,
    # outline
    outline_x0=None,
    outline_x1=None,
    # centers of all pixels
    cents_x=None,
    cents_y=None,
    cents_z=None,
    # unique vectors
    nin=None,
    e0=None,
    e1=None,
    # inwards normal vectors
    nin_x=None,
    nin_y=None,
    nin_z=None,
    # orthonormal direct base
    e0_x=None,
    e0_y=None,
    e0_z=None,
    e1_x=None,
    e1_y=None,
    e1_z=None,
):

    # ----
    # key

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('camera', {}), short='cam', key=key,
    )

    # ---------
    # outline

    outline_x0, outline_x1, area = _check_polygon_2d(
        poly_x=outline_x0,
        poly_y=outline_x1,
        poly_name=f'{key}-outline',
        can_be_None=False,
        closed=False,
        counter_clockwise=True,
        return_area=True,
    )

    # -----------
    # cents

    cents_x = np.atleast_1d(cents_x).ravel().astype(float)
    cents_y = np.atleast_1d(cents_y).ravel().astype(float)
    cents_z = np.atleast_1d(cents_z).ravel().astype(float)

    # shapes
    if not (cents_x.shape == cents_y.shape == cents_z.shape):
        lstr = [
            ('cents_x', cents_x.shape),
            ('cents_y', cents_y.shape),
            ('cents_z', cents_z.shape),
        ]
        lstr = [f"\t- {kk}.shape: {vv}" for kk, vv in lstr]
        msg = (
            "Args cents_x, cents_y, cents_z must have the same shape!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    iout = ~(
        np.isfinite(cents_x) & np.isfinite(cents_y) & np.isfinite(cents_z)
    )
    if np.any(iout):
        msg = (
            f"Non-finite cents detected for cam1d '{key}':\n"
            f"\t- indices: {iout.nonzero()[0]}\n"
            f"\t- cents_x: {cents_x[iout]}\n"
            f"\t- cents_y: {cents_y[iout]}\n"
            f"\t- cents_z: {cents_z[iout]}\n"
        )
        raise Exception(msg)

    # total nb of pixels
    npix = cents_x.size

    # make sure all cents are different
    dist = np.full((npix,), np.nan)
    for ii in range(npix):
        dist[:] = (
            (cents_x - cents_x[ii])**2
            + (cents_y - cents_y[ii])**2
            + (cents_z - cents_z[ii])**2
        )
        dist[ii] = 10
        if np.any(dist < 1.e-15):
            msg = (
                "Identical cents detected:\n"
                f"\t- ref: {ii}\n"
                f"\t- identicals: {(dist < 1.e-15).nonzero()[0]}\n"
            )
            raise Exception(msg)

    # -------------
    # unit vectors
    # -------------

    # ---------
    # check

    lc = [
        nin is not None,
        nin_x is not None,
    ]

    if np.sum(lc) != 1:
        msg = (
            "Please provide nin xor nin_x!"
        )
        raise Exception(msg)

    # ---------
    # unique

    if lc[0]:
        parallel = True
        nin, e0, e1 = ds._generic_check._check_vectbasis(
            e0=nin,
            e1=e0,
            e2=e1,
            dim=3,
        )

    else:

        lv = [
            ['nin_x', nin_x], ['nin_y', nin_y], ['nin_z', nin_z],
            ['e0_x', e0_x], ['e0_y', e0_y], ['e0_z', e0_z],
            ['e1_x', e1_x], ['e1_y', e1_y], ['e1_z', e1_z],
        ]

        # check they are all provided
        lNone = [vv[0] for vv in lv if vv[1] is None]
        if len(lNone) > 0:
            msg = (
                f"All unit vectors must be provided for camera '{key}'!\n"
                f"The following are not provided: {lNone}"
            )
            raise Exception(msg)

        # particular case: scalar because common to all
        c0 = all([
            np.isscalar(vv[1]) or np.array(vv[1]).size == 1 for vv in lv
        ])
        if c0:
            parallel = True
            nin, e0, e1 = ds._generic_check._check_vectbasis(
                e0=np.r_[nin_x, nin_y, nin_z],
                e1=np.r_[e0_x, e0_y, e0_z],
                e2=np.r_[e1_x, e1_y, e1_z],
                dim=3,
            )

        else:

            parallel = False

            # force into numpy array
            for vv in lv:
                vv[1] = np.atleast_1d(vv[1]).ravel().astype(float)

            # check shapes
            dshape = {
                vv[0]: vv[1].shape for vv in lv if vv[1].shape != (npix,)
            }
            if len(set(dshape.values())) > 1:
                lstr = [f"\t- {k0}: {v0}" for k0, v0 in dshape.items()]
                msg = (
                    f"All unit vector componant must have shape ({npix},)!\n"
                    + "\n".join(lstr)
                )
                raise Exception(msg)

            # force normalization
            norm = np.sqrt((lv[0][1]**2 + lv[1][1]**2 + lv[2][1]**2))
            nin_x = lv[0][1] / norm
            nin_y = lv[1][1] / norm
            nin_z = lv[2][1] / norm

            norm = np.sqrt((lv[3][1]**2 + lv[4][1]**2 + lv[5][1]**2))
            e0_x = lv[3][1] / norm
            e0_y = lv[4][1] / norm
            e0_z = lv[5][1] / norm

            norm = np.sqrt((lv[6][1]**2 + lv[7][1]**2 + lv[8][1]**2))
            e1_x = lv[6][1] / norm
            e1_y = lv[7][1] / norm
            e1_z = lv[8][1] / norm

            # check perpendicularity
            sca = (nin_x*e0_x + nin_y*e0_y + nin_z*e0_z)
            if np.any(np.abs(sca) > 1e-14):
                msg = (
                    "Non-perpendicular nin vs e0:\n"
                    f"{(sca > 1.e-14).nonzero()[0]}"
                )
                raise Exception(msg)

            sca = (nin_x*e1_x + nin_y*e1_y + nin_z*e1_z)
            if np.any(np.abs(sca) > 1e-14):
                msg = (
                    "Non-perpendicular nin vs e1:\n"
                    f"{(sca > 1.e-14).nonzero()[0]}"
                )
                raise Exception(msg)

            sca = (e0_x*e1_x + e0_y*e1_y + e0_z*e1_z)
            if np.any(np.abs(sca) > 1e-14):
                msg = (
                    "Non-perpendicular e0 vs e1:\n"
                    f"{(sca > 1.e-14).nonzero()[0]}"
                )
                raise Exception(msg)

            # check right-handedness
            sca = (
                e1_x * (nin_y * e0_z - nin_z * e0_y)
                + e1_y * (nin_z * e0_x - nin_x * e0_z)
                + e1_z * (nin_x * e0_y - nin_y * e0_x)
            )
            if np.any(sca <= 0.):
                msg = (
                    "The following unit vectors do not seem right-handed:\n"
                    f"{(sca <= 0.).nonzero()[0]}"
                )
                raise Exception(msg)

            nin = (nin_x, nin_y, nin_z)
            e0 = (e0_x, e0_y, e0_z)
            e1 = (e1_x, e1_y, e1_z)

    return (
        key,
        outline_x0, outline_x1,
        area,
        cents_x,
        cents_y,
        cents_z,
        npix,
        parallel,
        nin, e0, e1,
    )


def _camera_1d(
    coll=None,
    key=None,
    # common 2d outline
    outline_x0=None,
    outline_x1=None,
    # centers of all pixels
    cents_x=None,
    cents_y=None,
    cents_z=None,
    # unique vectors
    nin=None,
    e0=None,
    e1=None,
    # inwards normal vectors
    nin_x=None,
    nin_y=None,
    nin_z=None,
    # orthonormal direct base
    e0_x=None,
    e0_y=None,
    e0_z=None,
    e1_x=None,
    e1_y=None,
    e1_z=None,
):

    # ------------
    # check inputs

    (
        key,
        outline_x0, outline_x1,
        area,
        cents_x,
        cents_y,
        cents_z,
        npix,
        parallel,
        nin, e0, e1,
    ) = _camera_1d_check(
        coll=coll,
        key=key,
        # outline
        outline_x0=outline_x0,
        outline_x1=outline_x1,
        # centers of all pixels
        cents_x=cents_x,
        cents_y=cents_y,
        cents_z=cents_z,
        # unique vectors
        nin=nin,
        e0=e0,
        e1=e1,
        # inwards normal vectors
        nin_x=nin_x,
        nin_y=nin_y,
        nin_z=nin_z,
        # orthonormal direct base
        e0_x=e0_x,
        e0_y=e0_y,
        e0_z=e0_z,
        e1_x=e1_x,
        e1_y=e1_y,
        e1_z=e1_z,
    )

    # ----------
    # dref

    npts = outline_x0.size
    knpts = f'{key}_npts'
    knpix = f'{key}_npix'
    dref = {
        knpts: {'size': npts},
        knpix: {'size': npix},
    }

    # -------------
    # ddata

    kcx = f'{key}_cx'
    kcy = f'{key}_cy'
    kcz = f'{key}_cz'
    kout0 = f'{key}_outx0'
    kout1 = f'{key}_outx1'

    ddata = {
        kout0: {
            'data': outline_x0,
            'ref': knpts,
            'dim': 'distance',
            'quant': 'x0',
            'name': 'x0',
            'units': 'm',
        },
        kout1: {
            'data': outline_x1,
            'ref': knpts,
            'dim': 'distance',
            'quant': 'x1',
            'name': 'x1',
            'units': 'm',
        },
        kcx: {
            'data': cents_x,
            'ref': knpix,
            'dim': 'distance',
            'quant': 'x',
            'name': 'x',
            'units': 'm',
        },
        kcy: {
            'data': cents_y,
            'ref': knpix,
            'dim': 'distance',
            'quant': 'y',
            'name': 'y',
            'units': 'm',
        },
        kcz: {
            'data': cents_z,
            'ref': knpix,
            'dim': 'distance',
            'quant': 'z',
            'name': 'z',
            'units': 'm',
        },
    }

    # -----
    # dobj

    if parallel:
        o_nin = nin
        o_e0 = e0
        o_e1 = e1

    else:
        kinx = f'{key}_nin_x'
        kiny = f'{key}_nin_y'
        kinz = f'{key}_nin_z'
        ke0x = f'{key}_e0_x'
        ke0y = f'{key}_e0_y'
        ke0z = f'{key}_e0_z'
        ke1x = f'{key}_e1_x'
        ke1y = f'{key}_e1_y'
        ke1z = f'{key}_e1_z'

        o_nin = (kinx, kiny, kinz)
        o_e0 = (ke0x, ke0y, ke0z)
        o_e1 = (ke1x, ke1y, ke1z)

    # dobj
    dobj = {
        'camera': {
            key: {
                'dgeom': {
                    'type': '',
                    'nd': '1d',
                    'parallel': parallel,
                    'shape': (npix,),
                    'ref': (knpix,),
                    'ref_flat': (knpix,),
                    'pix_area': area,
                    'pix_nb': npix,
                    'outline': (kout0, kout1),
                    'cent': None,
                    'cents': (kcx, kcy, kcz),
                    'nin': o_nin,
                    'e0': o_e0,
                    'e1': o_e1,
                },
            },
        },
    }

    # ------------------------
    # parallel vs non-parallel

    if not parallel:
        ddata.update({
            kinx: {
                'data': nin[0],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'x',
                'name': 'x',
                'units': 'm',
            },
            kiny: {
                'data': nin[1],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'y',
                'name': 'y',
                'units': 'm',
            },
            kinz: {
                'data': nin[2],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'z',
                'name': 'z',
                'units': 'm',
            },
            ke0x: {
                'data': e0[0],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'x',
                'name': 'x',
                'units': 'm',
            },
            ke0y: {
                'data': e0[1],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'y',
                'name': 'y',
                'units': 'm',
            },
            ke0z: {
                'data': e0[2],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'z',
                'name': 'z',
                'units': 'm',
            },
            ke1x: {
                'data': e1[0],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'x',
                'name': 'x',
                'units': 'm',
            },
            ke1y: {
                'data': e1[1],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'y',
                'name': 'y',
                'units': 'm',
            },
            ke1z: {
                'data': e1[2],
                'ref': knpix,
                'dim': 'distance',
                'quant': 'z',
                'name': 'z',
                'units': 'm',
            },
        })

    return dref, ddata, dobj


# ##################################################################
# ##################################################################
#                           Camera 2d
# ##################################################################


def _camera_2d_check(
    coll=None,
    key=None,
    # outline
    outline_x0=None,
    outline_x1=None,
    # centers of all pixels
    cent=None,
    cents_x0=None,
    cents_x1=None,
    # inwards normal vectors
    nin=None,
    e0=None,
    e1=None,
):

    # ----
    # key

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('camera', {}), short='cam', key=key,
    )

    # ---------
    # outline

    outline_x0, outline_x1, area = _check_polygon_2d(
        poly_x=outline_x0,
        poly_y=outline_x1,
        poly_name=f"'{key}' (outline_x0, outline_x1)",
        can_be_None=False,
        closed=False,
        counter_clockwise=True,
        return_area=True,
    )

    # -----------
    # cent

    cent = np.atleast_1d(cent).ravel().astype(float)
    if cent.shape != (3,) or np.any(~np.isfinite(cent)):
        msg = f"Arg cent non valid shape {cent.shape} vs (3,) or non-finite!"
        raise Exception(msg)

    # -----------
    # cents

    cents_x0 = np.atleast_1d(cents_x0).ravel().astype(float)
    cents_x1 = np.atleast_1d(cents_x1).ravel().astype(float)

    # finite
    iout = ~np.isfinite(cents_x0)
    if np.any(iout):
        msg = "Non-finite cents_x0 detected:\n{iout.nonzero()[0]}"
        raise Exception(msg)

    iout = ~np.isfinite(cents_x1)
    if np.any(iout):
        msg = "Non-finite cents_x1 detected:\n{iout.nonzero()[0]}"
        raise Exception(msg)

    # total nb of pixels
    npix0 = cents_x0.size
    npix1 = cents_x1.size

    # make sure all cents are different
    if np.unique(cents_x0).size != cents_x0.size:
        msg = "Double values found in cents_x0!"
        raise Exception(msg)

    if np.unique(cents_x1).size != cents_x1.size:
        msg = "Double values found in cents_x1!"
        raise Exception(msg)

    # -----------
    # unit vectors

    lv = [('nin', nin), ('e0', e0), ('e1', e1)]

    # check they are all provided
    lNone = [vv[0] for vv in lv if vv[1] is None]
    if len(lNone) > 0:
        msg = (
            f"All unit vectors must be provided for camera '{key}'!\n"
            f"The following are not provided: {lNone}"
        )
        raise Exception(msg)

    # particular case: scalar because common to all
    nin, e0, e1 = ds._generic_check._check_vectbasis(
        e0=nin,
        e1=e0,
        e2=e1,
        dim=3,
    )

    # extenthalf for compatibility with other surfaces
    extenthalf = [
        0.5*(outline_x0.max() - outline_x0.min()),
        0.5*(outline_x1.max() - outline_x1.min()),
    ]

    return (
        key,
        outline_x0, outline_x1,
        area,
        cent,
        cents_x0, cents_x1,
        npix0, npix1,
        nin, e0, e1,
        extenthalf,
    )


def _camera_2d(
    coll=None,
    key=None,
    # common 2d outline
    outline_x0=None,
    outline_x1=None,
    # centers of all pixels
    cent=None,
    cents_x0=None,
    cents_x1=None,
    # inwards normal vectors
    nin=None,
    e0=None,
    e1=None,
):

    # ------------
    # check inputs

    (
        key,
        outline_x0, outline_x1,
        area,
        cent,
        cents_x0, cents_x1,
        npix0, npix1,
        nin, e0, e1,
        extenthalf,
    ) = _camera_2d_check(
        coll=coll,
        key=key,
        # outline
        outline_x0=outline_x0,
        outline_x1=outline_x1,
        # centers of all pixels
        cent=cent,
        cents_x0=cents_x0,
        cents_x1=cents_x1,
        # inwards normal vectors
        nin=nin,
        e0=e0,
        e1=e1,
    )

    # ----------
    # dref

    npts = outline_x0.size
    knpts = f'{key}_npts'
    knpix0 = f'{key}_npix0'
    knpix1 = f'{key}_npix1'
    knpix = f'{key}_npix'
    dref = {
        knpts: {'size': npts},
        knpix0: {'size': npix0},
        knpix1: {'size': npix1},
        knpix: {'size': npix0*npix1},
    }

    # -------------
    # ddata

    kc0 = f'{key}_c0'
    kc1 = f'{key}_c1'
    kout0 = f'{key}_outx0'
    kout1 = f'{key}_outx1'

    ddata = {
        kout0: {
            'data': outline_x0,
            'ref': knpts,
            'dim': 'distance',
            'quant': 'x0',
            'name': 'x0',
            'units': 'm',
        },
        kout1: {
            'data': outline_x1,
            'ref': knpts,
            'dim': 'distance',
            'quant': 'x1',
            'name': 'x1',
            'units': 'm',
        },
        kc0: {
            'data': cents_x0,
            'ref': knpix0,
            'dim': 'distance',
            'quant': 'x0',
            'name': 'x0',
            'units': 'm',
        },
        kc1: {
            'data': cents_x1,
            'ref': knpix1,
            'dim': 'distance',
            'quant': 'x1',
            'name': 'x1',
            'units': 'm',
        },
    }

    # -----
    # dobj

    dobj = {
        'camera': {
            key: {
                'dgeom': {
                    'type': 'planar',
                    'nd': '2d',
                    'parallel': True,
                    'shape': (npix0, npix1),
                    'ref': (knpix0, knpix1),
                    'ref_flat': (knpix,),
                    'pix_area': area,
                    'pix_nb': npix0 * npix1,
                    'outline': (kout0, kout1),
                    'cent': cent,
                    'cents': (kc0, kc1),
                    'nin': nin,
                    'e0': e0,
                    'e1': e1,
                    'extenthalf': extenthalf,
                },
            },
        },
    }

    return dref, ddata, dobj


# ##################################################################
# ##################################################################
#                       Camera dmat
# ##################################################################


def _dmat(
    coll=None,
    key=None,
    dmat=None,
):
    """ Check qeff is provided as a 1d vector vs lamb or energy

    lamb is assumed to be in m and converted to energy
    energy is assumed to be in eV

    """

    # ------------
    # trivial case

    if dmat is None:
        return None, None, None

    dref, ddata = {}, {}

    # ------
    # Check

    # Check dict type and content (each key is a valid string)
    dmat = ds._generic_check._check_dict_valid_keys(
        var=dmat,
        varname='dmat',
        has_all_keys=False,
        has_only_keys=True,
        keys_can_be_None=None,
        dkeys=_DMAT_KEYS,
        return_copy=True,
    )

    if dmat.get('bins') is not None:
        dmat['mode'] = 'PHA'

    # -----------------
    # check PHA vs bins

    if dmat['mode'] == 'PHA':
        if dmat.get('bins') is None:
            msg = (
                f"The bins (eV) must be provided for camera {key} in mode PHA!"
            )
            raise Exception(msg)

        kb = f'{key}_bin'
        coll.add_bins(
            key=kb,
            edges=dmat['bins'],
            units='eV',
            quant='E',
            dim='energy',
        )
        dmat['bins'] = kb

    # -----------------------------------
    # check energy / qeff values

    if all([dmat.get(kk) is not None for kk in ['energy', 'qeff']]):

        dmat['qeff_E'], dmat['qeff'] = _class4_check._dmat_energy_trans(
            energ=dmat['qeff_E'],
            trans=dmat['qeff'],
        )

        # ----------
        # dref

        kne = f'{key}_qnE'
        ne = dmat['qeff_E'].size
        dref[kne] = {'size': ne}

        # ----------
        # ddata

        kqE = f'{key}_qE'
        kqeff = f'{key}_qeff'

        ddata.update({
            kqE: {
                'data': dmat['qeff_E'],
                'ref': kne,
                'dim': 'energy',
                'quant': 'E',
                'name': 'E',
                'units': 'eV',
            },
            kqeff: {
                'data': dmat['qeff'],
                'ref': kne,
                'dim': None,
                'quant': 'quantum eff.',
                'name': '',
                'units': '',
            },
        })

        # -----------
        # dmat

        dmat['qeff_E'] = kqE
        dmat['qeff'] = kqeff

    return dref, ddata, dmat


# ##################################################################
# ##################################################################
#                           Utilities
# ##################################################################


def get_camera_unitvectors(
    coll=None,
    key=None,
):

    # ---------
    # check key

    lok = list(coll.dobj.get('camera', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # ---------------------------
    # get unit vector components

    dgeom = coll.dobj['camera'][key]['dgeom']
    if dgeom['parallel']:
        dout = {
            'nin_x': dgeom['nin'][0],
            'nin_y': dgeom['nin'][1],
            'nin_z': dgeom['nin'][2],
            'e0_x':  dgeom['e0'][0],
            'e0_y':  dgeom['e0'][1],
            'e0_z':  dgeom['e0'][2],
            'e1_x':  dgeom['e1'][0],
            'e1_y':  dgeom['e1'][1],
            'e1_z':  dgeom['e1'][2],
        }
    else:
        dout = {
            'nin_x': coll.ddata[dgeom['nin'][0]]['data'],
            'nin_y': coll.ddata[dgeom['nin'][1]]['data'],
            'nin_z': coll.ddata[dgeom['nin'][2]]['data'],
            'e0_x':  coll.ddata[dgeom['e0'][0]]['data'],
            'e0_y':  coll.ddata[dgeom['e0'][1]]['data'],
            'e0_z':  coll.ddata[dgeom['e0'][2]]['data'],
            'e1_x':  coll.ddata[dgeom['e1'][0]]['data'],
            'e1_y':  coll.ddata[dgeom['e1'][1]]['data'],
            'e1_z':  coll.ddata[dgeom['e1'][2]]['data'],
        }

    return dout


def get_camera_dxyz(coll=None, key=None, include_center=None):

    # ---------
    # check key

    lok = [
        k0 for k0, v0 in coll.dobj.get('camera', {}).items()
        if v0['dgeom']['parallel'] is True
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    dgeom = coll.dobj['camera'][key]['dgeom']

    # include_center
    include_center = ds._generic_check._check_var(
        include_center, 'include_center',
        types=bool,
        default=True,
    )

    # ----------------
    # get unit vectors

    e0 = dgeom['e0']
    e1 = dgeom['e1']

    out0 = coll.ddata[dgeom['outline'][0]]['data']
    out1 = coll.ddata[dgeom['outline'][1]]['data']

    if include_center is True:
        out0 = np.append(0, out0)
        out1 = np.append(0, out1)

    dx = out0 * e0[0] + out1 * e1[0]
    dy = out0 * e0[1] + out1 * e1[1]
    dz = out0 * e0[2] + out1 * e1[2]

    return dx, dy, dz


def get_camera_cents_xyz(coll=None, key=None):

    # ---------
    # check key

    lok = list(coll.dobj.get('camera', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # ---------------------------
    # get unit vector components

    dgeom = coll.dobj['camera'][key]['dgeom']
    if dgeom['nd'] == '1d':
        cx = coll.ddata[dgeom['cents'][0]]['data']
        cy = coll.ddata[dgeom['cents'][1]]['data']
        cz = coll.ddata[dgeom['cents'][2]]['data']
    else:
        c0 = coll.ddata[dgeom['cents'][0]]['data']
        c1 = coll.ddata[dgeom['cents'][1]]['data']

        cx = (
            dgeom['cent'][0]
            + np.repeat(c0[:, None], c1.size, axis=1) * dgeom['e0'][0]
            + np.repeat(c1[None, :], c0.size, axis=0) * dgeom['e1'][0]
        )
        cy = (
            dgeom['cent'][1]
            + np.repeat(c0[:, None], c1.size, axis=1) * dgeom['e0'][1]
            + np.repeat(c1[None, :], c0.size, axis=0) * dgeom['e1'][1]
        )
        cz = (
            dgeom['cent'][2]
            + np.repeat(c0[:, None], c1.size, axis=1) * dgeom['e0'][2]
            + np.repeat(c1[None, :], c0.size, axis=0) * dgeom['e1'][2]
        )

    return cx, cy, cz


import os
import sys
import time
import shutil
import datetime
import tempfile
import platform
import subprocess

from distutils.dist import Distribution
from distutils.sysconfig import customize_compiler
from numpy.distutils.ccompiler import new_compiler
from numpy.distutils.command.config_compiler import config_cc


def is_platform_windows():
    """Checks if platfom is windows

    Return
    --------
    bool
        True if on Windows
        False if not

    """
    return platform.system() == "Windows"


omp_source = """
#include <omp.h>
#include <stdio.h>
int main(void) {
  #pragma omp parallel
  printf("nthreads=%d\\n", omp_get_num_threads());
  return 0;
}
"""


def get_compiler():
    """Get a compiler equivalent to the one that will be used to build tofu
    when doing setup.py

    Handles compiler specified as follows:
        - python setup.py build_ext --compiler=<compiler>
        - CC=<compiler> python setup.py build_ext
    """
    dist = Distribution({'script_name': os.path.basename(sys.argv[0]),
                         'script_args': sys.argv[1:],
                         'cmdclass': {'config_cc': config_cc}})
    dist.parse_config_files()
    dist.parse_command_line()

    cmd_opts = dist.command_options.get('build_ext')

    if cmd_opts is not None and 'compiler' in cmd_opts:
        compiler = cmd_opts['compiler'][1]
    else:
        compiler = None

    ccompiler = new_compiler(compiler=compiler)
    customize_compiler(ccompiler)
    return ccompiler


def get_openmp_flag(compiler):
    """Returns list of flags for using OpenMP depending on compiler and
    platform.

    Parameters
    ----------
    compiler : numpy.distutils.compiler
        Compiler used when invoking setup.py build

    """
    if hasattr(compiler, 'compiler'):
        compiler = compiler.compiler[0]
    else:
        compiler = compiler.__class__.__name__

    if sys.platform == "win32" and ('icc' in compiler or 'icl' in compiler):
        return ['/Qopenmp']
    elif sys.platform == "win32":
        return ['/openmp']
    elif sys.platform in ("darwin", "linux") and "icc" in compiler:
        return ['-qopenmp']
    elif sys.platform == "darwin" and 'openmp' in os.getenv('CPPFLAGS', ''):
        return ['-openmp']
    # Default flag for GCC and clang:
    return ['-fopenmp']


def check_for_openmp():
    """Compiles small code sample to see if OpenMP is installed or not.
    Returns corresponding flags to compile depending on platform and
    compiler.

    Returns
    -------
    int
        0 if no error, else subprocess.call error code
    list
        list of strings indicating openmp flags corresponding or empty
    """

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    filename = "test.c"
    compiler = get_compiler()
    flag_omp = get_openmp_flag(compiler)

    # getting compiler name
    if hasattr(compiler, 'compiler'):
        compiler = compiler.compiler[0]
    else:
        compiler = compiler.__class__.__name__

    try:
        with open(filename, "w") as file:
            file.write(omp_source)
        with open(os.devnull, "w") as fnull:
            result = subprocess.call(
                [compiler] + flag_omp + [filename], stdout=fnull, stderr=fnull,
                shell=is_platform_windows()
            )
    except subprocess.CalledProcessError:
        result = -1

    finally:
        # in any case, go back to previous cwd and clean up
        os.chdir(curdir)
        shutil.rmtree(tmpdir)
        if not result == 0:
            flag_omp = []
    return result, flag_omp


def is_openmp_installed():
    """ Returns ``True`` if OpenMP is installed, else ``False`` and a list of
    flags needed to compile with openmp if installed (or an empty list)
    and generates ``tofu.openmp_enabled.is_openmp_enabled``, which can then be
    used to determine, post build, whether the package was built with or
    without OpenMP support."""

    openmp_support, flag = check_for_openmp()
    openmp_enabled = not openmp_support
    generate_openmp_enabled_py(openmp_enabled)

    return openmp_enabled, flag


_IS_OPENMP_ENABLED_SRC = """
# Autogenerated by Tofu's setup.py on {timestamp!s}
def is_openmp_enabled():
    \"\"\"
    Determine whether this package was built with OpenMP support.
    \"\"\"
    return {return_bool}
"""[1:]


def generate_openmp_enabled_py(openmp_support, srcdir='.'):
    """
    Generate ``tofu.openmp_enabled.is_openmp_enabled``, which can then be used
    to determine, post build, whether the package was built with or without
    OpenMP support.
    """

    epoch = int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))
    timestamp = datetime.datetime.utcfromtimestamp(epoch)

    src = _IS_OPENMP_ENABLED_SRC.format(timestamp=timestamp,
                                        return_bool=openmp_support)

    package_srcdir = os.path.join(srcdir, "tofu", "geom")
    is_openmp_enabled_py = os.path.join(package_srcdir, 'openmp_enabled.py')
    with open(is_openmp_enabled_py, 'w') as f:
        f.write(src)




# Common
import numpy as np
import scipy.optimize as scpopt
import scipy.sparse as scpsparse
from scipy.interpolate import BSpline

# Temporary for debugging
import matplotlib.pyplot as plt


_JAC = 'dense'


###########################################################
###########################################################
#
#           Main function for fit1d
#
###########################################################
###########################################################


def multigausfit1d_from_dlines_funccostjac(
    lamb=None,
    indx=None,
    dinput=None,
    dind=None,
    jac=None,
):

    if jac is None:
        jac = _JAC

    ibckax = dind['bck_amp']['x'][0, :]
    ibckrx = dind['bck_rate']['x'][0, :]
    nbck = 1    # ibckax.size + ibckrx.size
    iax = dind['amp']['x'][0, :]
    iwx = dind['width']['x'][0, :]
    ishx = dind['shift']['x'][0, :]
    idratiox, idshx = None, None
    if dinput['double'] is not False:
        c0 = dinput['double'] is True
        if c0 or dinput['double'].get('dratio') is None:
            idratiox = dind['dratio']['x'][0, :]
        if c0 or dinput['double'].get('dshift') is None:
            idshx = dind['dshift']['x'][0, :]

    ial = dind['amp']['lines'][0, :]
    iwl = dind['width']['lines'][0, :]
    ishl = dind['shift']['lines'][0, :]

    iaj = dind['amp']['jac']
    iwj = dind['width']['jac']
    ishj = dind['shift']['jac']

    coefsal = dinput['amp']['coefs']
    coefswl = dinput['width']['coefs']
    coefssl = dinput['shift']['coefs']

    offsetal = dinput['amp']['offset']
    offsetwl = dinput['width']['offset']
    offsetsl = dinput['shift']['offset']

    lambrel = lamb - dinput['lambmin_bck']
    lambnorm = lamb[..., None]/dinput['lines'][None, ...]

    xscale = np.full((dind['sizex'],), np.nan)
    if indx is None:
        indx = np.ones((dind['sizex'],), dtype=bool)

    # func_details returns result in same shape as input
    def func_detail(
        x,
        xscale=xscale,
        indx=indx,
        lambrel=lambrel,
        lambnorm=lambnorm,
        ibckax=ibckax,
        ibckrx=ibckrx,
        ial=ial,
        iwl=iwl,
        ishl=ishl,
        idratiox=idratiox,
        idshx=idshx,
        nlines=dinput['nlines'],
        nbck=nbck,
        coefsal=coefsal[None, :],
        coefswl=coefswl[None, :],
        coefssl=coefssl[None, :],
        offsetal=offsetal[None, :],
        offsetwl=offsetwl[None, :],
        offsetsl=offsetsl[None, :],
        double=dinput['double'],
        scales=None,
        indok=None,
        const=None,
    ):
        if indok is None:
            indok = np.ones(lamb.shape, dtype=bool)
        shape = tuple(np.r_[indok.sum(), nbck+nlines])
        y = np.full(shape, np.nan)
        xscale[indx] = x*scales[indx]
        xscale[~indx] = const

        # Prepare
        amp = xscale[ial]*coefsal + offsetal
        wi2 = xscale[iwl]*coefswl + offsetwl
        shift = xscale[ishl]*coefssl + offsetsl
        exp = np.exp(-(lambnorm[indok, :] - (1 + shift))**2 / (2*wi2))

        if double is not False:
            # coefssl are line-specific, they do not affect dshift
            if double is True:
                dratio = xscale[idratiox]
                dshift = shift + xscale[idshx]  # *coefssl
            else:
                dratio = double.get('dratio', xscale[idratiox])
                dshift = shift + double.get('dshift', xscale[idshx])
            expd = np.exp(-(lambnorm[indok, :] - (1 + dshift))**2 / (2*wi2))

        # compute y
        y[:, :nbck] = (
            xscale[ibckax]
            * np.exp(xscale[ibckrx]*lambrel[indok])
        )[:, None]
        y[:, nbck:] = amp * exp
        if double is not False:
            y[:, nbck:] += amp * dratio * expd
        return y

    # cost and jacob return flattened results (for least_squares())
    def cost(
        x,
        xscale=xscale,
        indx=indx,
        lambrel=lambrel,
        lambnorm=lambnorm,
        ibckax=ibckax,
        ibckrx=ibckrx,
        ial=ial,
        iwl=iwl,
        ishl=ishl,
        idratiox=idratiox,
        idshx=idshx,
        scales=None,
        coefsal=coefsal[None, :],
        coefswl=coefswl[None, :],
        coefssl=coefssl[None, :],
        offsetal=offsetal[None, :],
        offsetwl=offsetwl[None, :],
        offsetsl=offsetsl[None, :],
        double=dinput['double'],
        indok=None,
        const=None,
        data=0.,
    ):
        if indok is None:
            indok = np.ones(lambrel.shape, dtype=bool)

        # xscale = x*scales   !!! scales ??? !!! TBC
        xscale[indx] = x*scales[indx]
        xscale[~indx] = const

        # make sure iwl is 2D to get all lines at once
        amp = xscale[ial] * coefsal + offsetal
        inv_2wi2 = 1./(2.*(xscale[iwl] * coefswl + offsetwl))
        shift = xscale[ishl] * coefssl + offsetsl
        y = (
            np.nansum(
                amp * np.exp(-(lambnorm[indok, :]-(1 + shift))**2 * inv_2wi2),
                axis=1,
            )
            + xscale[ibckax]*np.exp(xscale[ibckrx]*lambrel[indok])
        )
        if double is not False:
            if double is True:
                dratio = xscale[idratiox]
                # scales[ishl] or scales[idshx] ? coefssl ? +> no
                dshift = shift + xscale[idshx]
            else:
                dratio = double.get('dratio', xscale[idratiox])
                dshift = shift + double.get('dshift', xscale[idshx])

            y += np.nansum((amp * dratio
                            * np.exp(-(lambnorm[indok, :] - (1 + dshift))**2
                                     * inv_2wi2)), axis=1)

        if isinstance(data, np.ndarray):
            return y - data[indok]
        else:
            return y - data

    # Prepare jac
    if jac in ['dense', 'sparse']:
        # Define a callable jac returning (nlamb, sizex) matrix of partial
        # derivatives of np.sum(func_details(scaled), axis=0)

        # Pre-instanciate jacobian
        jac0 = np.zeros((lamb.size, dind['sizex']), dtype=float)
        # jac0 = np.zeros((lamb.size, indx.sum()), dtype=float)
        indxn = indx.nonzero()[0]

        def jacob(
            x,
            xscale=xscale,
            indx=indx,
            indxn=indxn,
            lambnorm=lambnorm,
            ibckax=ibckax,
            ibckrx=ibckrx,
            iax=iax, iaj=iaj, ial=ial,
            iwx=iwx, iwj=iwj, iwl=iwl,
            ishx=ishx, ishj=ishj, ishl=ishl,
            idratiox=idratiox, idshx=idshx,
            coefsal=coefsal[None, :],
            coefswl=coefswl[None, :],
            coefssl=coefssl[None, :],
            offsetal=offsetal[None, :],
            offsetwl=offsetwl[None, :],
            offsetsl=offsetsl[None, :],
            double=dinput['double'],
            scales=None, indok=None, data=None,
            jac0=jac0,
            const=None,
        ):
            """ Basic docstr """
            if indok is None:
                indok = np.ones(lamb.shape, dtype=bool)
            xscale[indx] = x*scales[indx]
            xscale[~indx] = const

            # Intermediates
            amp = xscale[ial] * coefsal + offsetal
            wi2 = xscale[iwl] * coefswl + offsetwl
            inv_wi2 = 1./wi2
            shift = xscale[ishl] * coefssl + offsetsl
            beta = (lambnorm[indok, :] - (1 + shift)) * inv_wi2 / 2.
            alpha = -beta**2 * (2*wi2)
            exp = np.exp(alpha)

            # Background
            jac0[indok, ibckax[0]] = (
                scales[ibckax[0]]
                * np.exp(xscale[ibckrx]*lambrel[indok])
            )
            jac0[indok, ibckrx[0]] = (
                xscale[ibckax[0]] * scales[ibckrx[0]] * lambrel[indok]
                * np.exp(xscale[ibckrx]*lambrel[indok])
            )

            # amp (shape: nphi/lamb, namp[jj])
            # for jj in range(len(iaj)):
            for jj, aa in enumerate(iaj):
                jac0[indok, iax[jj]] = np.sum(
                    exp[:, aa] * coefsal[:, aa],
                    axis=1) * scales[iax[jj]]

            # width2
            for jj in range(len(iwj)):
                jac0[indok, iwx[jj]] = np.sum(
                    (-alpha[:, iwj[jj]] * amp[:, iwj[jj]]
                     * exp[:, iwj[jj]] * coefswl[:, iwj[jj]]
                     * inv_wi2[:, iwj[jj]]), axis=1) * scales[iwx[jj]]

            # shift
            for jj in range(len(ishj)):
                jac0[indok, ishx[jj]] = np.sum(
                    (amp[:, ishj[jj]] * 2. * beta[:, ishj[jj]])
                    * exp[:, ishj[jj]] * coefssl[:, ishj[jj]],
                    axis=1) * scales[ishx[jj]]

            # double
            if double is not False:

                if double is True:
                    dratio = xscale[idratiox]
                    # coefssl are line-specific, they do not affect dshift
                    dshift = shift + xscale[idshx]
                else:
                    dratio = double.get('dratio', xscale[idratiox])
                    dshift = shift + double.get('dshift', xscale[idshx])
                # ampd not defined to save memory => *dratio instead
                # ampd = amp*dratio
                betad = (lambnorm[indok, :] - (1 + dshift)) * inv_wi2 / 2.
                alphad = -betad**2 * (2*wi2)
                expd = np.exp(alphad)

                # amp
                for jj in range(len(iaj)):
                    jac0[indok, iax[jj]] += dratio*np.sum(
                        expd[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1) * scales[iax[jj]]

                # width2
                for jj in range(len(iwj)):
                    jac0[indok, iwx[jj]] += np.sum(
                        (
                            -alphad[:, iwj[jj]] * amp[:, iwj[jj]]
                            * expd[:, iwj[jj]] * coefswl[:, iwj[jj]]
                            * inv_wi2[:, iwj[jj]]
                        ),
                        axis=1,
                    ) * scales[iwx[jj]] * dratio

                # shift
                for jj in range(len(ishj)):
                    jac0[indok, ishx[jj]] += np.sum(
                        (amp[:, ishj[jj]] * 2.*betad[:, ishj[jj]]
                         * expd[:, ishj[jj]] * coefssl[:, ishj[jj]]),
                        axis=1) * scales[ishx[jj]] * dratio

                # dratio
                if double is True or double.get('dratio') is None:
                    jac0[indok, idratiox] = (
                        scales[idratiox] * np.sum(amp * expd, axis=1)
                    )

                # dshift
                if double is True or double.get('dshift') is None:
                    jac0[indok, idshx] = dratio * np.sum(
                        amp * 2.*betad*scales[idshx] * expd, axis=1)

            return jac0[indok, :][:, indx]

    elif jac == 'sparse':
        msg = "Sparse jacobian is pointless for 1d spectrum fitting"
        raise Exception(msg)

    if jac not in ['dense', 'sparse', 'LinearSparseOperator']:
        if jac not in ['2-point', '3-point']:
            msg = ("jac should be in "
                   + "['dense', 'sparse', 'LinearsparseOp', "
                   + "'2-point', '3-point']\n"
                   + "\t- provided: {}".format(jac))
            raise Exception(msg)
        jacob = jac

    return func_detail, cost, jacob

###########################################################
###########################################################
#
#           Main function for fit2d
#
###########################################################
###########################################################


def multigausfit2d_from_dlines_funccostjac(
    phi_flat=None,
    dinput=None,
    dind=None,
    jac=None,
):

    return_costjac = phi_flat is not None

    ibckax = dind['bck_amp']['x']
    ibckrx = dind['bck_rate']['x']
    nbck = 1    # ibckax.size + ibckrx.size
    iax = dind['amp']['x']
    iwx = dind['width']['x']
    ishx = dind['shift']['x']
    idratiox, idshx = None, None
    if dinput['double'] is not False:
        c0 = dinput['double'] is True
        if c0 or dinput['double'].get('dratio') is None:
            idratiox = dind['dratio']['x']
        if c0 or dinput['double'].get('dshift') is None:
            idshx = dind['dshift']['x']

    ial = dind['amp']['lines']
    iwl = dind['width']['lines']
    ishl = dind['shift']['lines']

    iaj = dind['amp']['jac']
    iwj = dind['width']['jac']
    ishj = dind['shift']['jac']

    coefsal = dinput['amp']['coefs'][None, :]
    coefswl = dinput['width']['coefs'][None, :]
    coefssl = dinput['shift']['coefs'][None, :]

    offsetal = dinput['amp']['offset'][None, :]
    offsetwl = dinput['width']['offset'][None, :]
    offsetsl = dinput['shift']['offset'][None, :]

    xscale = np.full((dind['sizex'],), np.nan)

    km = dinput['knots_mult']
    kpb = dinput['nknotsperbs']
    nbs = dinput['nbs']

    BS = BSpline(
        km,
        np.ones(ial.shape, dtype=float),
        dinput['deg'],
        extrapolate=False,
        axis=0,
    )

    nlines = dinput['nlines']
    deg = dinput['deg']
    double = dinput['double']
    lambmin_bck = dinput['lambmin_bck']

    # Pre-set kwdargs
    dkwdargs = dict(
        ibckax=ibckax,
        ibckrx=ibckrx,
        ial=ial,
        iwl=iwl,
        ishl=ishl,
        idratiox=idratiox,
        idshx=idshx,
        nlines=nlines,
        nbck=nbck,
        km=km,
        kpb=kpb,
        nbs=nbs,
        deg=deg,
        BS=BS,
        coefsal=coefsal[None, :],
        coefswl=coefswl[None, :],
        coefssl=coefssl[None, :],
        offsetal=offsetal[None, :],
        offsetwl=offsetwl[None, :],
        offsetsl=offsetsl[None, :],
        double=double,
        lambmin_bck=lambmin_bck,
    )

    # ----------------------------
    # details and sum for end-user

    # func_details returns result in same shape as input
    def func_detail(
        x,
        xscale=xscale,
        phi=None,
        lamb=None,
        scales=None,
        ind_bs=None,
        const=None,
        indx=None,
        **dkwdargs,
    ):
        # normalize lamb
        lambrel = lamb - lambmin_bck
        shape = tuple(np.r_[[1 for ii in range(lamb.ndim)], -1])
        lambn = lamb[..., None] / dinput['lines'].reshape(shape)

        # shape = tuple(np.r_[indok.sum(), nbck+nlines, nbs])
        shape = tuple(np.r_[phi.shape, nbck + nlines, nbs])
        y = np.full(shape, np.nan)

        if indx is None:
            xscale = x*scales
        else:
            xscale[indx] = x*scales[indx]
            xscale[~indx] = const

        # bck rate
        BS.c = xscale[ibckrx]
        bckr = BS(phi)

        # make sure iwl is 2D to get all lines at once
        BS.c = xscale[iwl] * coefswl + offsetwl
        wi2 = BS(phi)
        BS.c = xscale[ishl] * coefssl + offsetsl
        shift = BS(phi)
        exp = np.exp(-(lambn - (1 + shift))**2 / (2*wi2))

        if double is not False:
            # coefssl are line-specific, they do not affect dshift
            if double is True:
                dratio = xscale[idratiox[:, 0]]
                dshift = shift + xscale[idshx[:, 0]]  # *coefssl
            else:
                dratio = double.get('dratio', xscale[idratiox[:, 0]])
                dshift = shift + double.get('dshift', xscale[idshx[:, 0]])
            expd = np.exp(-(lambn - (1 + dshift))**2 / (2*wi2))

        # Loop on individual bsplines for amp
        for ii in range(nbs):

            if ind_bs is not None and ii not in ind_bs:
                continue

            bs = BSpline.basis_element(
                km[ii:ii + kpb],
                extrapolate=False,
            )(phi)

            indbs = np.isfinite(bs)
            if not np.any(indbs):
                continue
            bs = bs[indbs]

            # bck
            y[indbs, 0, ii] = (
                xscale[ibckax[ii, 0]]*bs
                * np.exp(bckr[indbs, 0]*lambrel[indbs])
            )

            # lines
            for jj in range(nlines):
                amp = (
                    bs * xscale[ial[ii, jj]] * coefsal[:, jj] + offsetal[:, jj]
                )
                y[indbs, nbck+jj, ii] = amp * exp[indbs, jj]
                if double is not False:
                    y[indbs, nbck+jj, ii] += (amp * dratio * expd[indbs, jj])

        return y

    # cost and jacob return flattened results (for least_squares())
    def func_sum(
        x,
        xscale=xscale,
        phi=None,
        lamb=None,
        scales=None,
        ind_bs=None,
        const=None,
        indx=None,
        **dkwdargs,
    ):

        # normalize lamb
        lambrel = lamb - lambmin_bck
        shape = tuple(np.r_[[1 for ii in range(lamb.ndim)], -1])
        lambn = lamb[..., None] / dinput['lines'].reshape(shape)

        # xscale = x*scales
        if indx is None:
            xscale = x*scales
        else:
            xscale[indx] = x*scales[indx]
            xscale[~indx] = const

        # Background
        BS.c = xscale[ibckax][:, 0]
        bcka = BS(phi)
        BS.c = xscale[ibckrx][:, 0]
        y = bcka * np.exp(BS(phi)*lambrel)

        # make sure iwl is 2D to get all lines at once
        BS.c = xscale[ial] * coefsal + offsetal
        amp = BS(phi)
        BS.c = xscale[iwl] * coefswl + offsetwl
        wi2 = BS(phi)
        BS.c = xscale[ishl] * coefssl + offsetsl
        csh = BS(phi)

        y += np.nansum(
            amp * np.exp(-(lambn - (1 + csh))**2 / (2*wi2)),
            axis=-1,
        )

        if double is not False:
            if double is True:
                dratio = xscale[idratiox[:, 0]]
                # scales[ishl] or scales[idshx] ? coefssl ? +> no
                dcsh = csh + xscale[idshx[:, 0]]
            else:
                dratio = double.get('dratio', xscale[idratiox[:, 0]])
                dcsh = csh + double.get('dshift', xscale[idshx[:, 0]])

            expd = np.exp(-(lambn - (1 + dcsh))**2 / (2*wi2))
            y += np.nansum(amp * dratio * expd, axis=-1)
        return y

    # -------------------------------
    # cost and jacob for optimization

    func_cost, func_jacob = None, None
    if return_costjac:

        def func_cost(
            x,
            xscale=xscale,
            scales=None,
            indok_flat=None,
            ind_bs=None,
            const=None,
            indx=None,
            data_flat=None,
            phi_flat=None,
            lambrel_flat=None,
            lambn_flat=None,
            jac0=None,
            libs=None,
            **dkwdargs,
        ):

            assert data_flat.shape == lambrel_flat.shape
            assert indok_flat is not None

            if indx is None:
                xscale = x*scales
            else:
                xscale[indx] = x*scales[indx]
                xscale[~indx] = const

            # Background
            BS.c = xscale[ibckax][:, 0]
            bcka = BS(phi_flat)
            BS.c = xscale[ibckrx][:, 0]
            y = bcka * np.exp(BS(phi_flat)*lambrel_flat)

            # make sure iwl is 2D to get all lines at once
            BS.c = xscale[ial] * coefsal + offsetal
            amp = BS(phi_flat)
            BS.c = xscale[iwl] * coefswl + offsetwl
            wi2 = BS(phi_flat)
            BS.c = xscale[ishl] * coefssl + offsetsl
            csh = BS(phi_flat)

            y += np.nansum(
                amp * np.exp(-(lambn_flat - (1 + csh))**2 / (2*wi2)),
                axis=-1,
            )

            if double is not False:
                if double is True:
                    dratio = xscale[idratiox[:, 0]]
                    # scales[ishl] or scales[idshx] ? coefssl ? +> no
                    dcsh = csh + xscale[idshx[:, 0]]
                else:
                    dratio = double.get('dratio', xscale[idratiox[:, 0]])
                    dcsh = csh + double.get('dshift', xscale[idshx[:, 0]])

                expd = np.exp(-(lambn_flat - (1 + dcsh))**2 / (2*wi2))
                y += np.nansum(amp * dratio * expd, axis=-1)

            return y - data_flat

        def func_jacob(
            x,
            xscale=xscale,
            scales=None,
            indok_flat=None,
            ind_bs=None,
            const=None,
            indx=None,
            data_flat=0.,
            phi_flat=None,
            lambrel_flat=None,
            lambn_flat=None,
            jac0=None,
            libs=None,
            **dkwdargs,
        ):
            """ Basic docstr """
            if indx is None:
                xscale = x*scales
            else:
                xscale[indx] = x*scales[indx]
                xscale[~indx] = const

            # Intermediates

            # Loop on bs
            for ii in range(nbs):

                # phi interval
                ibs = libs[ii]

                # bspline
                bs = BSpline.basis_element(
                    km[ii:ii+kpb],
                    extrapolate=False,
                )(phi_flat[ibs])[:, None]

                # check bspline
                if np.any(~np.isfinite(bs)):
                    msg = "Non-finite values in bs (affecting jacobian)!"
                    raise Exception(msg)

                # Intermediates - background
                BS.c = xscale[ibckax[:, 0]]
                bcka = BS(phi_flat[ibs])
                BS.c[...] = xscale[ibckrx[:, 0]]
                bckr = BS(phi_flat[ibs])
                expbck = np.exp(bckr*lambrel_flat[ibs])

                # Intermediates - amp, wi2, shift
                BS.c = xscale[ial] * coefsal + offsetal
                amp = BS(phi_flat[ibs])
                BS.c[...] = xscale[iwl] * coefswl + offsetwl
                wi2 = BS(phi_flat[ibs])
                BS.c[...] = xscale[ishl] * coefssl + offsetsl
                shift = BS(phi_flat[ibs])

                beta = (lambn_flat[ibs, :] - (1 + shift)) / (2*wi2)
                alpha = -beta**2 * (2*wi2)
                # exp = np.exp(alpha)
                bsexp = bs * np.exp(alpha)

                # Background amplitude
                jac0[ibs, ibckax[ii, 0]] = (
                    bs[:, 0] * scales[ibckax[ii, 0]] * expbck
                )

                # Background rate
                jac0[ibs, ibckrx[ii, 0]] = (
                    bs[:, 0] * scales[ibckrx[ii, 0]]
                    * lambrel_flat[ibs] * bcka * expbck
                )

                # amp (shape: nphi/lamb, namp[jj])
                for jj in range(len(iaj)):
                    ix = iax[ii, jj]
                    jac0[ibs, ix] = np.sum(
                        bsexp[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1,
                    ) * scales[ix]

                # width2
                for jj in range(len(iwj)):
                    ix = iwx[ii, jj]
                    jac0[ibs, ix] = np.sum(
                        (
                            -alpha[:, iwj[jj]] * amp[:, iwj[jj]]
                            * bsexp[:, iwj[jj]] * coefswl[:, iwj[jj]]
                            / wi2[:, iwj[jj]]
                        ),
                        axis=1,
                    ) * scales[ix]

                # shift
                for jj in range(len(ishj)):
                    ix = ishx[ii, jj]
                    jac0[ibs, ix] = np.sum(
                        (
                            amp[:, ishj[jj]] * 2. * beta[:, ishj[jj]]
                            * bsexp[:, ishj[jj]] * coefssl[:, ishj[jj]]
                        ),
                        axis=1,
                    ) * scales[ix]

                # double
                if double is False:
                    continue

                if double is True:
                    dratio = xscale[idratiox[:, 0]]
                    # coefssl are line-specific, they do not affect dshift
                    dshift = shift + xscale[idshx[:, 0]]
                else:
                    dratio = double.get('dratio', xscale[idratiox[:, 0]])
                    dshift = shift + double.get('dshift', xscale[idshx[:, 0]])

                # ampd = amp*dratio
                betad = (lambn_flat[ibs, :] - (1 + dshift)) / (2*wi2)
                alphad = -betad**2 * (2*wi2)
                expd = np.exp(alphad)
                bsexpd = bs * expd

                # amp
                for jj in range(len(iaj)):
                    ix = iax[ii, jj]
                    jac0[ibs, ix] += dratio*np.sum(
                        bsexpd[:, iaj[jj]] * coefsal[:, iaj[jj]],
                        axis=1,
                    ) * scales[ix]

                # width2
                for jj in range(len(iwj)):
                    ix = iwx[ii, jj]
                    jac0[ibs, ix] += np.sum(
                        (-alphad[:, iwj[jj]] * amp[:, iwj[jj]]
                         * bsexpd[:, iwj[jj]] * coefswl[:, iwj[jj]]
                         / wi2[:, iwj[jj]]),
                        axis=1,
                    ) * scales[ix] * dratio

                # shift
                for jj in range(len(ishj)):
                    ix = ishx[ii, jj]
                    jac0[ibs, ix] += np.sum(
                        (amp[:, ishj[jj]] * 2.*betad[:, ishj[jj]]
                         * bsexpd[:, ishj[jj]] * coefssl[:, ishj[jj]]),
                        axis=1,
                    ) * scales[ix] * dratio

                # dratio
                if double is True or double.get('dratio') is None:
                    jac0[ibs, idratiox[:, 0]] = (
                        scales[idratiox[:, 0]] * np.sum(amp * expd, axis=1)
                    )

                # dshift
                if double is True or double.get('dshift') is None:
                    jac0[ibs, idshx[:, 0]] = dratio * np.sum(
                        amp * 2.*betad*scales[idshx] * expd,
                        axis=1,
                    )
            if indx is None:
                return jac0
            else:
                return jac0[:, indx]

    return func_detail, func_sum, func_cost, func_jacob



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 30 17:55:54 2024

@author: dvezinet
"""


import os
import itertools as itt


import numpy as np
import datastock as ds


from . import _class08_save2json
from . import _class08_loadfromjson
from . import _class08_save2stp


# #################################################################
# #################################################################
#          Default values
# #################################################################


_NAME = 'test'


# #################################################################
# #################################################################
#          save
# #################################################################


def save(
    # ---------------
    # input from tofu
    coll=None,
    key=None,
    key_cam=None,
    key_optics=None,
    # ---------------
    # options
    factor=None,
    color=None,
    empty_name=None,
    # ---------------
    # saving
    pfe_save=None,
    overwrite=None,
):

    # ----------------
    #   check
    # ----------------

    (
        key, key_cam,
        factor, color,
        pfe_save, overwrite,
    ) = _check_save(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # options
        factor=factor,
        color=color,
        # saving
        pfe_save=pfe_save,
        overwrite=overwrite,
        ext=None,
    )

    # ----------------
    #   call
    # ----------------

    if pfe_save.endswith('.json'):
        _class08_save2json.main(
            coll=coll,
            key=key,
            key_cam=key_cam,
            # options
            factor=factor,
            color=color,
            # saving
            pfe_save=pfe_save,
            overwrite=overwrite,
        )

    elif pfe_save.endswith('.stp'):
        _class08_save2stp.main(
            # input from tofu
            coll=coll,
            key=key,
            key_cam=key_cam,
            key_optics=key_optics,
            # options
            factor=factor,
            color=color,
            empty_name=empty_name,
            # saving
            pfe_save=pfe_save,
            overwrite=overwrite,
        )

    return


# #################################################################
# #################################################################
#          load
# #################################################################


def load(
    pfe=None,
    coll=None,
    returnas=None,
):

    # ----------------
    #   check
    # ----------------

    coll, pfe, fname = _check_load(
        coll=coll,
        pfe=pfe,
    )

    # ----------------
    #   call
    # ----------------

    if pfe.endswith('.json'):
        return _class08_loadfromjson.main(
            coll=coll,
            pfe=pfe,
            fname=fname,
            returnas=returnas,
        )


# #################################################################
# #################################################################
#          check - save
# #################################################################


def _check_save(
    coll=None,
    key=None,
    key_cam=None,
    key_optics=None,
    # options
    factor=None,
    color=None,
    # saving
    pfe_save=None,
    overwrite=None,
    ext='json',
):

    # ---------------
    # ext
    # ---------------

    lok = ['json', 'stp']
    ext = ds._generic_check._check_var(
        ext, 'ext',
        types=str,
        default='json',
        allowed=lok,
    )

    # ---------------
    # pfe_save
    # ---------------

    # Default
    if pfe_save is None:
        path = os.path.abspath('.')
        name = key if key is not None else _NAME
        pfe_save = os.path.join(path, f"{name}.{ext}")

    # check
    c0 = (
        isinstance(pfe_save, str)
        and (
            os.path.split(pfe_save)[0] == ''
            or os.path.isdir(os.path.split(pfe_save)[0])
        )
    )
    if not c0:
        msg = (
            f"Arg pfe_save must be a saving file str ending in '.{ext}'!\n"
            f"Provided: {pfe_save}"
        )
        raise Exception(msg)

    # makesure extension is included
    ext_file = pfe_save.split('.')[-1]
    if ext_file not in lok:
        pfe_save = f"{pfe_save}.{ext}"

    # ---------------
    # factor
    # ---------------

    factor = float(ds._generic_check._check_var(
        factor, 'factor',
        types=(float, int),
        default=1.,
    ))

    # ----------------
    # overwrite
    # ----------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    # ---------------
    # key
    # ---------------

    if ext == 'json':
        key, key_cam = coll.get_diagnostic_cam(
            key=key,
            key_cam=key_cam,
            default='all',
        )

    return (
        key, key_cam,
        factor, color,
        pfe_save, overwrite,
    )


# #################################################################
# #################################################################
#          check
# #################################################################


def _check_load(
    coll=None,
    pfe=None,
):

    # ---------------
    # coll
    # ---------------

    from ._class10_Inversion import Inversion as Collection

    if coll is None:
        coll = Collection()

    else:
        c0 = isinstance(coll, Collection)
        if not c0:
            msg = (
                "Arg coll must be a tf.data.Collection instance!\n"
                f"\t Provided:\n{coll}\n"
            )
            raise Exception(msg)

    # ---------------
    # pfe
    # ---------------

    c0 = (
        isinstance(pfe, str)
        and os.path.isfile(pfe)
        and pfe.endswith('.json')
    )
    if not c0:
        msg = (
            "Arg 'pfe' must be a 'path/file.ext' to an existing json file!\n"
            f"\t- Provided: {pfe}\n"
        )
        raise Exception(msg)

    # -------------
    # fname
    # -------------

    fname = os.path.split(pfe)[-1][:-4]

    return coll, pfe, fname




import copy


# ################################################
# ################################################
#              IDS Names
# ################################################


_DIDS = {
    "summary": 'sum',
    'equilibrium': 'eq',
    'core_profiles': 'cprof',
}


# ################################################
# ################################################
#              SHORTS
# ################################################


_DSHORT = {

    # ---------------
    # summary
    # ---------------

    "summary": {

        # -----------
        # time traces

        't': {
            'dim': 'time',
            'long': 'time',
            'units': 's',
            'ref0': 'nt',
        },
        'Q': {
            'dim': 'gain',
            'long': 'global_quantities.fusion_gain.value[nt]',
            'units': None,
        },
        'H98': {
            'dim': None,
            'long': 'global_quantities.h_98.value[nt]',
            'units': None,
        },
    },

    # ---------------
    # equilibrium
    # ---------------

    "equilibrium": {

        # -----------
        # time traces

        't': {
            'dim': 'time',
            'long': 'time',
            'units': 's',
            'ref0': 'nt',
        },
        'Ip': {
            'dim': 'current',
            'long': 'time_slice[nt].global_quantities.ip',
            'units': 'A',
        },

        # --------
        # mesh 2d

        'm2dR': {
            'dim': 'distance',
            'long': 'time_slice[0].profiles_2d[im2d].grid.dim1',
            'units': 'm',
            'mesh': 'm2d',
        },
        'm2dZ': {
            'dim': 'distance',
            'long': 'time_slice[0].profiles_2d[im2d].grid.dim2',
            'units': 'm',
            'mesh': 'm2d',
        },

        # ----------
        # 2d data

        '2dpsi': {
            'dim': 'B flux',
            'name': 'psi',
            'long': 'time_slice[nt].profiles_2d[im2d].psi',
            'units': 'Wb',
        },
        '2dphi': {
            'dim': 'B flux',
            'name': 'phi',
            'long': 'time_slice[nt].profiles_2d[im2d].phi',
            'units': 'Wb',
        },
        '2dBR': {
            'dim': 'B',
            'name': 'BR',
            'long': 'time_slice[nt].profiles_2d[im2d].b_field_r',
            'units': 'T',
        },
        '2dBZ': {
            'dim': 'B',
            'name': 'BZ',
            'long': 'time_slice[nt].profiles_2d[im2d].b_field_z',
            'units': 'T',
        },
        '2dBphi': {
            'dim': 'B',
            'name': 'Btor',
            'long': 'time_slice[nt].profiles_2d[im2d].b_field_tor',
            'units': 'T',
        },

        # ----------
        # 1d data

        '1ddpdpsi': {
            'dim': 'p / flux',
            'long': 'time_slice[nt].profiles_1d.dpressure_dpsi[im1d]',
            'units': 'Pa/Wb',
        },
        '1ddVdpsi': {
            'dim': 'vol / flux',
            'long': 'time_slice[nt].profiles_1d.dvolume_dpsi[im1d]',
            'units': 'm3/Wb',
        },
        '1dkappa': {
            'dim': 'adim',
            'long': 'time_slice[nt].profiles_1d.elongation[im1d]',
            'units': None,
        },
        '1djtor': {
            'dim': 'current density',
            'long': 'time_slice[nt].profiles_1d.j_tor[im1d]',
            'units': 'A/m2',
        },
        '1dp': {
            'dim': 'pressure',
            'long': 'time_slice[nt].profiles_1d.pressure[im1d]',
            'units': 'Pa',
        },
        '1dpsi': {
            'dim': 'B flux',
            'name': 'psi',
            'long': 'time_slice[nt].profiles_1d.psi[im1d]',
            'units': 'Wb',
        },
        '1dpsin': {
            'dim': 'B flux norm',
            'name': 'psin',
            'long': 'time_slice[nt].profiles_1d.psi_norm[im1d]',
            'units': None,
        },
        '1dphi': {
            'dim': 'B flux',
            'name': 'phi',
            'long': 'time_slice[nt].profiles_1d.phi[im1d]',
            'units': 'Wb',
        },
        '1dq': {
            'dim': 'q',
            'long': 'time_slice[nt].profiles_1d.q[im1d]',
            'units': None,
        },
        '1drin': {
            'dim': 'distance',
            'description': 'LCFS major radius, inboard',
            'long': 'time_slice[nt].profiles_1d.r_inboard[im1d]',
            'units': 'm',
        },
        '1drout': {
            'dim': 'distance',
            'description': 'LCFS major radius, outboard',
            'long': 'time_slice[nt].profiles_1d.r_outboard[im1d]',
            'units': 'm',
        },
        '1drhot': {
            'dim': 'rho',
            'name': 'rhot',
            'long': 'time_slice[nt].profiles_1d.rho_tor[im1d]',
            'units': None,
        },
        '1drhotn': {
            'dim': 'rho',
            'name': 'rhotn',
            'long': 'time_slice[nt].profiles_1d.rho_tor_norm[im1d]',
            'units': None,
        },
        '1dS': {
            'dim': 'surface',
            'long': 'time_slice[nt].profiles_1d.surface[im1d]',
            'units': 'm2',
        },
        '1dtrianglow': {
            'dim': 'triangularity',
            'long': 'time_slice[nt].profiles_1d.triangularity_lower[im1d]',
            'units': None,
        },
        '1dtriangup': {
            'dim': 'triangularity',
            'long': 'time_slice[nt].profiles_1d.triangularity_upper[im1d]',
            'units': None,
        },
        '1dV': {
            'dim': 'volume',
            'long': 'time_slice[nt].profiles_1d.volume[im1d]',
            'units': 'm3',
        },
    },

    # ---------------
    # core_profiles
    # ---------------

    "core_profiles": {

        # -----------
        # time traces

        't': {
            'dim': 'time',
            'long': 'time',
            'units': 's',
            'ref0': 'nt',
        },
        'Vloop': {
            'dim': 'voltage',
            'name': 'Vloop',
            'long': 'global_quantities.v_loop[nt]',
            'units': 'V',
        },
        'TeVol': {
            'dim': 'Te',
            'name': 'Te',
            'long': 'global_quantities.t_e_volume_average[nt]',
            'units': 'eV',
        },
        'neVol': {
            'dim': 'density',
            'name': 'ne',
            'long': 'global_quantities.n_e_volume_average[nt]',
            'units': '1/m3',
        },
        'Ip': {
            'dim': 'current',
            'name': 'Ip',
            'long': 'global_quantities.ip[nt]',
            'units': 'A',
        },
        'betap': {
            'dim': 'beta',
            'name': 'betap',
            'long': 'global_quantities.beta_pol[nt]',
            'units': None,
        },
        'betat': {
            'dim': 'beta',
            'name': 'betat',
            'long': 'global_quantities.beta_tor[nt]',
            'units': None,
        },
        'betatn': {
            'dim': 'beta',
            'name': 'betatn',
            'long': 'global_quantities.beta_tor_norm[nt]',
            'units': None,
        },
        'Wdia': {
            'dim': 'Energy',
            'name': 'Wdia',
            'long': 'global_quantities.energy_diamagnetic[nt]',
            'units': 'J',
        },

        # -----------
        # profiles 1d

        '1dpsi': {
            'dim': 'B flux',
            'name': 'psi',
            'long': 'profiles_1d[nt].grid.psi[im1d]',
            'units': 'Wb',
        },
        '1dphi': {
            'dim': 'B flux',
            'name': 'phi',
            'long': 'profiles_1d[nt].grid.phi[im1d]',
            'units': 'Wb',
        },
        '1drhotn': {
            'dim': 'rho',
            'name': 'rhotn',
            'long': 'profiles_1d[nt].grid.rho_tor_norm[im1d]',
            'units': None,
        },
        '1drhopn': {
            'dim': 'rho',
            'name': 'rhopn',
            'long': 'profiles_1d[nt].grid.rho_pol_norm[im1d]',
            'units': None,
        },
        '1dV': {
            'dim': 'rho',
            'name': 'rhopn',
            'long': 'profiles_1d[nt].grid.rho_pol_norm[im1d]',
            'units': None,
        },
    },
}


# ###########################################
# ###########################################
#            get_dshort
# ###########################################


def get_dshort():

    dshort = copy.deepcopy(_DSHORT)

    for ids, vids in dshort.items():

        # ------------------
        # list of references

        lkref0 = [
            k0 for k0, v0 in vids.items()
            if v0.get('ref0') is not None
        ]
        lref0 = [dshort[ids][k0]['ref0'] for k0 in lkref0]

        for k0 in lkref0:
            dshort[ids][k0]['ref'] = (dshort[ids][k0]['ref0'],)

        # ---------------
        # list of meshes

        # lm = sorted(set([
        #     v0['mesh'] for k0, v0 in vids.items()
        #     if v0.get('mesh') is not None
        # ]))

        # -------------
        # list if data

        ldata = [
            k0 for k0, v0 in vids.items()
            if all([v0.get(ss) is None for ss in ['ref0', 'mesh']])
        ]

        # ---------
        # add ref

        for i0, k0 in enumerate(ldata):

            ref = []
            long = vids[k0]['long']
            ss = str(long)
            nr = long.count('[')

            for ir in range(nr):

                kr = ss[ss.index('[')+1:ss.index(']')]

                if kr.isnumeric():
                    continue

                if kr in lref0 + ['im1d', 'im2d']:
                    ref.append(kr)

                ss = ss[ss.index(']')+1:]

            dshort[ids][k0]['ref'] = tuple(ref)

    return dshort



# -*- coding: utf-8 -*-
"""
Created on Thu Jan 25 19:03:27 2024

@author: dvezinet
"""

# Built-in


# Common
import numpy as np
import scipy.integrate as scpinteg
import astropy.units as asunits


# #############################################################################
# #############################################################################
#                           LOS
# #############################################################################


def _compute_los(
    coll=None,
    key=None,
    key_bs=None,
    key_diag=None,
    key_cam=None,
    # sampling
    indbs=None,
    res=None,
    mode=None,
    key_integrand=None,
    radius_max=None,
    is3d=None,
    # common ref
    ref_com=None,
    ref_vector_strategy=None,
    # slicing
    shape_mat=None,
    sli_mat=None,
    axis_pix=None,
    # parameters
    brightness=None,
    verb=None,
):

    # -----
    # units

    units = asunits.m
    units_coefs = asunits.Unit()

    # ----------------
    # loop on cameras

    dout = {}
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    for k0 in key_cam:

        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        key_los = doptics[k0]['los']
        key_mat = f'{key}_{k0}'

        sh = tuple([npix if ss is None else ss for ss in shape_mat])
        mat = np.zeros(sh, dtype=float)

        # -----------------------
        # loop on group of pixels (to limit memory footprint)

        anyok = False
        for ii in range(npix):

            # verb
            if verb is True:
                msg = (
                    f"\t- '{key_mat}' for cam '{k0}': pixel {ii + 1} / {npix}"
                    f"\t{(mat > 0).sum()} / {mat.size}\t\t"
                )
                end = '\n' if ii == npix - 1 else '\r'
                print(msg, flush=True, end=end)

            # sample los
            out_sample = coll.sample_rays(
                key=key_los,
                res=res,
                mode=mode,
                segment=None,
                ind_ch=ii,
                radius_max=radius_max,
                concatenate=False,
                return_coords=['R', 'z', 'ltot'],
            )

            if out_sample is None or out_sample[0] is None:
                continue

            R, Z, length = out_sample

            # -------------
            # interpolate

            # datai, units, refi = coll.interpolate(
            douti = coll.interpolate(
                keys=None,
                ref_key=key_bs,
                # interpolation pts
                x0=R[:, 0],
                x1=Z[:, 0],
                submesh=True,
                grid=False,
                # common ref
                ref_com=ref_com,
                ref_vector_strategy=ref_vector_strategy,
                # bsplines-specific
                # azone=None,
                indbs_tf=indbs,
                details=True,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )[f'{key_bs}_details']

            datai, refi = douti['data'], douti['ref']
            axis = refi.index(None)
            iok = np.isfinite(datai)

            if not np.any(iok):
                continue

            datai[~iok] = 0.

            # ------------
            # integrate

            # check and update slice
            assert datai.ndim in [2, 3], datai.shape
            sli_mat[axis_pix] = ii

            # integrate
            mat[tuple(sli_mat)] = scpinteg.simpson(
                datai,
                x=length[:, 0],
                axis=axis,
            )

            anyok = True

        # --------------
        # post-treatment

        if anyok:
            # brightness
            if brightness is False:
                ketend = doptics[k0]['etendue']
                units_coefs = coll.ddata[ketend]['units']
                etend = coll.ddata[ketend]['data']
                sh_etend = [-1 if aa == axis else 1 for aa in range(len(refi))]
                mat *= etend.reshape(sh_etend)

            # set ref
            refi = list(refi)
            refi[axis] = coll.dobj['camera'][k0]['dgeom']['ref_flat']
            refi = tuple(np.r_[refi[:axis], refi[axis], refi[axis+1:]])

        else:
            refi = None
            axis = None

        # fill dout
        dout[key_mat] = {
            'data': mat,
            'ref': refi,
            'units': units * units_coefs,
        }

    return dout, axis


# #############################################################################
# #############################################################################
#                           VOS
# #############################################################################


def _compute_vos(
    coll=None,
    key=None,
    key_bs=None,
    key_diag=None,
    key_cam=None,
    # dvos
    dvos=None,
    # sampling
    indbs=None,
    res=None,
    mode=None,
    key_integrand=None,
    radius_max=None,
    is3d=None,
    # common ref
    ref_com=None,
    ref_vector_strategy=None,
    # slicing
    shape_mat=None,
    sli_mat=None,
    axis_pix=None,
    # parameters
    brightness=None,
    verb=None,
):

    # -----
    # units

    units = asunits.Unit(dvos[key_cam[0]]['sang_cross']['units'])
    units_coefs = asunits.Unit()

    # -------------
    # mesh sampling

    lkm = list(set([v0['keym'] for v0 in dvos.values()]))
    lres = list(set([tuple(v0['res_RZ']) for v0 in dvos.values()]))

    if len(lkm) != 1:
        lstr = [f"\t- '{k0}': '{v0['keym']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same mesh!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    if len(lres) != 1:
        lstr = [f"\t- '{k0}': '{v0['res_RZ']}'" for k0, v0 in dvos.items()]
        msg = (
            "All cameras vos were not sampled using the same resolution!\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # extract
    keym = lkm[0]
    res_RZ = list(lres[0])

    # mesh sampling
    dsamp = coll.get_sample_mesh(
        key=keym,
        res=res_RZ,
        mode='abs',
        grid=False,
        in_mesh=True,
        # non-used
        x0=None,
        x1=None,
        Dx0=None,
        Dx1=None,
        imshow=False,
        store=False,
        kx0=None,
        kx1=None,
    )

    x0u = dsamp['x0']['data']
    x1u = dsamp['x1']['data']

    # ----------------
    # loop on cameras

    dout = {}
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    for k0 in key_cam:

        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        key_mat = f'{key}_{k0}'

        # -------------
        # slicing

        is2d = coll.dobj['camera'][k0]['dgeom']['nd'] == '2d'
        if is2d:
            n0, n1 = coll.dobj['camera'][k0]['dgeom']['shape']
            sli = lambda ii: (ii // n1, ii % n1, slice(None))
        else:
            sli = lambda ii: (ii, slice(None))

        # shape, key
        sh = tuple([npix if ss is None else ss for ss in shape_mat])
        mat = np.zeros(sh, dtype=float)

        # ---------------------------------------------------
        # loop on group of pixels (to limit memory footprint)

        anyok = False
        for ii in range(npix):

            # verb
            if verb is True:
                msg = (
                    f"\t- '{key_mat}' for cam '{k0}': pixel {ii + 1} / {npix}"
                    f"\t{(mat > 0).sum()} / {mat.size}\t\t"
                )
                end = '\n' if ii == npix - 1 else '\r'
                print(msg, flush=True, end=end)

            # sample los
            indok = np.isfinite(dvos[k0]['sang_cross']['data'][sli(ii)])
            if not np.any(indok):
                continue

            # indices + dv
            indr = dvos[k0]['indr_cross']['data'][sli(ii)][indok]
            indz = dvos[k0]['indz_cross']['data'][sli(ii)][indok]

            # -------------
            # interpolate

            # datai, units, refi = coll.interpolate(
            douti = coll.interpolate(
                keys=None,
                ref_key=key_bs,
                x0=x0u[indr],
                x1=x1u[indz],
                submesh=True,
                grid=False,
                # common ref
                ref_com=ref_com,
                ref_vector_strategy=ref_vector_strategy,
                # bsplines-specific
                # azone=None,
                indbs_tf=indbs,
                details=True,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )[f'{key_bs}_details']

            datai, refi = douti['data'], douti['ref']
            axis = refi.index(None)
            iok = np.isfinite(datai)

            if not np.any(iok):
                continue

            datai[~iok] = 0.

            # ------------
            # integrate

            # check and update slice
            assert datai.ndim in [2, 3], datai.shape
            sli_mat[axis_pix] = ii

            # integrate
            mat[tuple(sli_mat)] = np.sum(
                datai * dvos[k0]['sang_cross']['data'][sli(ii)][indok][:, None],
                axis=axis,
            )

            anyok = True

        # --------------
        # post-treatment

        if anyok:
            # brightness
            if brightness is True:
                ketend = doptics[k0]['etendue']
                units_coefs = coll.ddata[ketend]['units']
                etend = coll.ddata[ketend]['data']
                sh_etend = [-1 if aa == axis else 1 for aa in range(len(refi))]
                mat /= etend.reshape(sh_etend)

            # set ref
            refi = list(refi)
            refi[axis] = coll.dobj['camera'][k0]['dgeom']['ref_flat']
            refi = tuple(np.r_[refi[:axis], refi[axis], refi[axis+1:]])

        else:
            refi = None
            axis = None

        # fill dout
        dout[key_mat] = {
            'data': mat,
            'ref': refi,
            'units': units / units_coefs,
        }

    return dout, axis